/***************************************************************************
 *
 * COPYRIGHTHERE
 *
 * $Id: plug.c,v 1.63.2.16 2003/11/03 14:11:14 sasa Exp $
 *
 * Author: Balazs Scheidler <bazsi@balabit.hu>
 * Auditor:
 * Last audited version:
 * Notes:
 *
 ***************************************************************************/

#include <zorp/plugsession.h>
#include <zorp/thread.h>
#include <zorp/streamfd.h>
#include <zorp/proxy.h>
#include <zorp/poll.h>
#include <zorp/policy.h>
#include <zorp/thread.h>
#include <zorp/log.h>
#include <zorp/registry.h>
#include <zorp/sockaddr.h>
#include <zorp/io.h>
#include <zorp/fastpath.h>

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <arpa/inet.h>

#define PLUG_DEFAULT_BUFSIZE    65535

#define PLUG_DEBUG "plug.debug"
#define PLUG_ERROR "plug.error"
#define PLUG_POLICY "plug.policy"
#define PLUG_SESSION "plug.session"

#define ZS_MATCH_SADDR 0x0001
#define ZS_MATCH_SPORT 0x0002
#define ZS_MATCH_DADDR 0x0004
#define ZS_MATCH_DPORT 0x0008

#define ZS_MATCH_ALL   0x000F

typedef struct _PlugProxy
{
  ZProxy super;
  ZSockAddr *master_client, *master_server;
  ZPoll *poll;
  GAsyncQueue *session_queue;
  ZPlugSessionData session_data;
  ZDispatchEntry *secondary_dispatch;
  guint secondary_mask;
  guint secondary_sessions;
} PlugProxy;

#define ENABLE_PKT_STATS 1

static gboolean
plug_packet_stat_event(ZProxy *s, ZSessionVars *vars, 
                       guint64 client_bytes, guint64 client_pkts, 
                       guint64 server_bytes, guint64 server_pkts);
static void plug_proxy_free(ZProxy *s);

void
plug_config_set_defaults(PlugProxy *self)
{
  z_proxy_enter(self);

  self->session_data.copy_to_server = TRUE;
  self->session_data.copy_to_client = TRUE;
  self->session_data.timeout = 600000;
  self->session_data.buffer_size = PLUG_DEFAULT_BUFSIZE;
  self->session_data.packet_stats = plug_packet_stat_event;
  
  /* secondary sessions are not permitted */
  self->secondary_mask = ZS_MATCH_ALL;
  self->secondary_sessions = 10;
  
  self->poll = z_poll_new();
  
  z_proxy_leave(self);
}

void
plug_register_vars(PlugProxy *self)
{
  z_proxy_enter(self);

  z_proxy_var_new(&self->super,
                  "secondary_mask",
                  Z_VAR_GET | Z_VAR_GET_CONFIG | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                  &self->secondary_mask);

  z_proxy_var_new(&self->super,
                  "secondary_sessions",
                  Z_VAR_GET | Z_VAR_GET_CONFIG | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                  &self->secondary_sessions);

  z_proxy_var_new(&self->super,
                  "timeout",
                  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                  &self->session_data.timeout);

  z_proxy_var_new(&self->super,
                  "copy_to_client",
                  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                  &self->session_data.copy_to_client);

  z_proxy_var_new(&self->super,
                  "copy_to_server",
                  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                  &self->session_data.copy_to_server);

  z_proxy_var_new(&self->super,
                  "shutdown_soft",
                  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                  &self->session_data.shutdown_soft);

  z_proxy_var_new(&self->super,
                  "packet_stats_interval_packet",
                  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                  &self->session_data.packet_stats_interval_packet);

  z_proxy_var_new(&self->super,
                  "packet_stats_interval_time",
                  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                  &self->session_data.packet_stats_interval_time);

  z_proxy_var_new(&self->super,
                  "buffer_size",
                  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                  &self->session_data.buffer_size);

  /* Zorp 1.4 compatibility */
  z_proxy_var_new(&self->super,
                  "packet_stats_interval",
                  Z_VAR_TYPE_ALIAS | Z_VAR_GET | Z_VAR_GET_CONFIG,
                  "packet_stats_interval_packet");

  z_proxy_leave(self);
}

static gboolean
plug_packet_stat_event(ZProxy *s, ZSessionVars *vars, 
                       guint64 client_bytes, guint64 client_pkts, 
                       guint64 server_bytes, guint64 server_pkts)
{
  PlugProxy *self = (PlugProxy *) s;
  ZPolicyObj *res;
  gboolean called;
  guint resc;

  z_policy_lock(self->super.thread);
  z_proxy_vars_set_active_session(s, vars);
  res = z_policy_call(self->super.handler, "packetStats",
                      z_policy_var_build("iiii",
                                         (guint32) client_bytes,
                                         (guint32) client_pkts,
                                         (guint32) server_bytes,
                                         (guint32) server_pkts),
                      &called,
                      self->super.session_id);
  z_proxy_vars_set_active_session(s, NULL);
  if (called)
    {
      resc = Z_REJECT;
      if (res)
        {
          if (!z_policy_var_parse(res, "i", &resc))
            z_proxy_log(self, PLUG_POLICY, 1, "Return value of packetStats() is not int;");
          else if (resc == Z_REJECT)
            z_proxy_log(self, PLUG_DEBUG, 6, "packetStats() return Z_REJECT, aborting session;");
        }
      else
        {
          z_proxy_log(self, PLUG_POLICY, 1, "Error during packetStats() event;");
        }
    }
  else
    resc = Z_ACCEPT;
  z_policy_var_unref(res);
  z_policy_unlock(self->super.thread);
  return resc == Z_ACCEPT;
}

static gboolean
plug_request_stack_event(PlugProxy *self, ZStackedProxy **stacked)
{
  ZPolicyObj *res;
  gboolean called;
  gboolean rc = TRUE;

  z_proxy_enter(self);
  z_policy_lock(self->super.thread);
  *stacked = NULL;
  res = z_policy_call(self->super.handler,
                      "requestStack",
                      NULL,
                      &called,
                      self->super.session_id);
  if (res)
    {
      if (res != z_policy_none)
        {
          *stacked = z_proxy_stack_proxy(&self->super, res);
        }
    }
  else if (called)
    res = FALSE;
    
  z_policy_var_unref(res);
  z_policy_unlock(self->super.thread);
  z_proxy_leave(self);
  return rc;
}

/* FIXME: make sure the proxy instance has not yet been freed yet, see
 * comment at plug_start_secondary_sessions */
static gboolean
plug_secondary_accept(ZConnection *conn, gpointer user_data)
{
  PlugProxy *self = (PlugProxy *) user_data;
  struct sockaddr_in *master_client, *master_server, *client, *server;
  
  if (!conn)
    {
      z_proxy_log(self, PLUG_ERROR, 3, "Error accepting secondary connection; conn='NULL'");
      return FALSE;
    }
    
  if (ZPROXY_GET_STATE(&self->super) != PS_WORKING)
    {
      return FALSE;
    }
  if (self->session_data.num_sessions >= self->secondary_sessions)
    {
      z_proxy_log(self, PLUG_DEBUG, 6, 
                  "Maximum number of secondary sessions reached; secondary_sessions='%d'", 
                  self->secondary_sessions);
      return FALSE;
    }
    
  if (!z_sockaddr_inet_check(conn->remote))
    {
      gchar buf[256];
      
      /* we are not interested in non-ipv4 secondary connections */
      z_proxy_log(self, PLUG_DEBUG, 6, 
                  "Incoming secondary connection with non-ipv4 client address; %s", 
                  z_connection_format(conn, buf, sizeof(buf)));
      return FALSE;
    }
    
  /* FIXME: put this comparison to the core somewhere, and make it
   * address independent
   */
  client = (struct sockaddr_in *) &conn->remote->sa;
  server = (struct sockaddr_in *) &conn->dest->sa;
  master_client = (struct sockaddr_in *) &self->master_client->sa;
  master_server = (struct sockaddr_in *) &self->master_server->sa;
  
  if ((self->secondary_mask & ZS_MATCH_SADDR) &&
      client->sin_addr.s_addr != master_client->sin_addr.s_addr)
    {
      return FALSE;
    }
  if ((self->secondary_mask & ZS_MATCH_SPORT) &&
      client->sin_port != master_client->sin_port)
    {
      return FALSE;
    }

  if ((self->secondary_mask & ZS_MATCH_DADDR) &&
      server->sin_addr.s_addr != master_server->sin_addr.s_addr)
    {
      return FALSE;
    }

  if ((self->secondary_mask & ZS_MATCH_DPORT) &&
      server->sin_port != master_server->sin_port)
    {
      return FALSE;
    }
  
  g_async_queue_push(self->session_queue, conn);
  z_poll_wakeup(self->poll);
  return TRUE;
}

static gboolean
plug_enable_secondary_sessions(PlugProxy *self)
{
  ZSockAddr *listen;
  ZDispatchParams params;
  guint protocol;
  
  z_proxy_enter(self);

  if (self->secondary_mask == ZS_MATCH_ALL)
    {
      z_proxy_leave(self);
      return TRUE;
    }

  if (!z_proxy_get_addresses(&self->super, &protocol, &self->master_client, &self->master_server, NULL, NULL, &listen))
    {
      z_proxy_log(self, PLUG_POLICY, 1, "Error starting secondary sessions, no listen address set;");
      z_proxy_leave(self);
      return FALSE;
    }
  
  /* we increment the refcount of self, which will be freed when the
   * dispatch entry is unregistered
   */
   
  z_proxy_ref(&self->super);
  
  memset(&params, 0, sizeof(params));
  self->secondary_dispatch = 
    z_dispatch_register(self->super.session_id, 
                        protocol, listen, NULL, 
                        ZD_PRI_NORMAL, 
                        &params, 
                        plug_secondary_accept, self, (GDestroyNotify) z_proxy_unref);
  z_sockaddr_unref(listen);
  
  z_proxy_leave(self);
  return TRUE;
}

static gboolean
plug_disable_secondary_sessions(PlugProxy *self)
{
  if (self->secondary_dispatch)
    {
      z_dispatch_unregister(self->secondary_dispatch);
      self->secondary_dispatch = NULL;
    }
  return TRUE;
}

static gboolean
plug_start_secondary_session(PlugProxy *self, ZConnection *conn)
{
  ZStream *server_stream;
  ZPlugSession *session;
  ZStackedProxy *stacked;
  gchar buf1[128], buf2[128];
  gint fd;
  
  fd = z_stream_get_fd(conn->stream);
  
  z_proxy_log(self, PLUG_SESSION, 3, 
              "Secondary session, client connection accepted; client_fd='%d', client_address='%s', client_local='%s'", 
              fd, z_sockaddr_format(conn->remote, buf1, sizeof(buf1)), z_sockaddr_format(conn->dest, buf2, sizeof(buf2)));
  
  server_stream = z_proxy_fastpath_connect_server_event(&self->super, conn);
  
  if (!server_stream)
    return FALSE;
  
  if (!plug_request_stack_event(self, &stacked))
    {
      z_stream_close(server_stream, NULL);
      z_stream_unref(server_stream);
      return FALSE;
    }
  
  session = z_plug_session_new(&self->super, &self->session_data, conn->stream, server_stream, stacked);
  if (!session)
    {
      z_stream_close(server_stream, NULL);
      z_stream_unref(server_stream);
      z_stacked_proxy_destroy(stacked);
      return FALSE;
    }
  z_stream_unref(server_stream);
    
  if (!z_plug_session_start(session, self->poll))
    {
      z_plug_session_free(session);
    }
  return TRUE;
}

static gboolean
plug_start_main_session(PlugProxy *self)
{
  ZPlugSession *session;
  
  z_proxy_enter(self);
  
  session = z_plug_session_new(&self->super, &self->session_data, self->super.endpoints[EP_CLIENT], self->super.endpoints[EP_SERVER], NULL);
  
  if (!session)
    {
      z_proxy_leave(self);
      return FALSE;
    }
  z_stream_unref(self->super.endpoints[EP_CLIENT]);
  z_stream_unref(self->super.endpoints[EP_SERVER]);
  self->super.endpoints[EP_CLIENT] = self->super.endpoints[EP_SERVER] = NULL;
  if (!z_plug_session_start(session, self->poll))
    {
      z_plug_session_free(session);
      z_proxy_leave(self);
      return FALSE;
    }
  z_proxy_leave(self);
  return TRUE;
}

static gboolean
plug_main_loop(PlugProxy *self)
{
  while ((g_async_queue_length(self->session_queue) || 
          z_plug_sessions_running(&self->session_data)) && 
         z_poll_iter_timeout(self->poll, -1))
    {
    
      while (g_async_queue_length(self->session_queue) > 0)
        {
          ZConnection *conn = (ZConnection *) g_async_queue_pop(self->session_queue);
          
          if (!plug_start_secondary_session(self, conn))
            z_connection_destroy(conn, TRUE);
          else
            z_connection_destroy(conn, FALSE);
        }
      z_plug_sessions_purge(&self->session_data);
    }
  z_plug_sessions_purge(&self->session_data);
  return TRUE;
}

static gboolean
plug_main(PlugProxy *self)
{
  z_proxy_enter(self);
  ZPROXY_SET_STATE(self, PS_CONFIG);
  if (!z_proxy_config_event(&self->super))
    {
      z_proxy_leave(self);
      return FALSE;
    }

  ZPROXY_SET_STATE(self, PS_STARTING_UP);
  if (!z_proxy_startup_event(&self->super))
    {
      z_proxy_leave(self);
      return FALSE;
    }

  if (!z_proxy_connect_server_event(&self->super, NULL, 0))
    {
      z_proxy_leave(self);
      return FALSE;
    }
    
  if (!plug_start_main_session(self))
    {
      z_proxy_leave(self);
      return FALSE;
    }
    
  if (!plug_enable_secondary_sessions(self))
    {
      z_proxy_leave(self);
      return FALSE;
    }

  ZPROXY_SET_STATE(self, PS_WORKING);
  plug_main_loop(self);
  
  /* disable secondary sessions */
  plug_disable_secondary_sessions(self);
  
  /* we might have some pending secondary connections which were accepted while
   * we switched into SHUTTING_DOWN state, handle those as well.
   */
  
  plug_main_loop(self);
  
  /* there might be no pending connections left */
  ZPROXY_SET_STATE(self, PS_SHUTTING_DOWN);
  z_proxy_shutdown_event(&self->super);
  
  z_proxy_leave(self);
  return TRUE;
}

static gpointer
plug_thread(gpointer s)
{
  PlugProxy *self = (PlugProxy *) s;

  z_proxy_enter(self);

  plug_config_set_defaults(self);
  plug_register_vars(self);
  plug_main(self);

  z_proxy_destroy(&self->super);
  z_leave();
  return NULL;
}

PlugProxy *
plug_init_proxy(gchar *session_id,
                ZStream *client,
                ZPolicyObj *handler)
{
  PlugProxy *self = g_new0(PlugProxy, 1);

  z_enter();
  z_proxy_init(&self->super, session_id, client, handler);
  
  self->super.free = plug_proxy_free;
  self->session_queue = g_async_queue_new();
  z_leave();
  return self;
}

ZProxy *
plug_proxy_new(gchar *session_id,
	       ZStream *client,
	       ZPolicyObj *handler)
{
  PlugProxy *self;
  
  z_enter();
  self = plug_init_proxy(session_id, client, handler);
  z_thread_new(session_id, plug_thread, self);
  z_leave();
  return &self->super;
}
    
static void
plug_proxy_free(ZProxy *s)
{
  PlugProxy *self = (PlugProxy *) s;
  
  z_proxy_enter(self);
  
  if (self->poll)
    {
      z_poll_unref(self->poll);
      self->poll = NULL;
    }
    

  if (self->session_queue)
    g_async_queue_unref(self->session_queue);

  z_sockaddr_unref(self->master_client);
  z_sockaddr_unref(self->master_server);
  z_proxy_leave(self);
}

gint
zorp_module_init(void)
{
  z_registry_add("plug", ZR_PROXY, plug_proxy_new);
  return TRUE;
}
